import unittest
from strmprivacy.diagnostics import PrivacyDiagnostics


class MyTestCase(unittest.TestCase):
    def test_happy_flow_1qi_1sa(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        d.calculate_stats(
            qi=['col3'],
            sa=['col1'],
            metrics=['k_anonymity', 'l_diversity', 't_closeness']
        )

        self.assertEqual(4, d.k, "k-anonymity for 'col3' is incorrect")
        self.assertEqual(2, d.l['col1'], "l-diversity for 'col1' is incorrect")
        self.assertAlmostEqual(0.23333, d.t, 4)

    def test_happy_flow_2qi_1sa(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        d.calculate_stats(
            qi=['col3', 'col4'],
            sa=['col1'],
            metrics=['k_anonymity', 'l_diversity', 't_closeness']
        )
        self.assertEqual(2, d.k, "k-anonymity for ['col3', 'col4'] is incorrect")
        self.assertEqual(2, d.l['col1'], "l-diversity for 'col1' is incorrect")
        self.assertAlmostEqual(0.43333, d.t, 4)

    def test_happy_flow_1qi_2sa(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        sa = ['col1', 'col2']
        d.calculate_stats(
            qi=['col4'],
            sa=sa,
            sa_types=['cat', 'num'],
            metrics=['k_anonymity', 'l_diversity', 't_closeness']
        )
        self.assertEqual(14, d.k, "k-anonymity for ['col3'] is incorrect")
        self.assertEqual(3, d.l['col1'], f"l-diversity for 'col1' is incorrect")
        self.assertEqual(14, d.l['col2'], f"l-diversity for 'col2' is incorrect")
        self.assertAlmostEqual(0.05238, d.t, 4)

    def test_non_existing_qi(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        with self.assertRaises(AssertionError) as context:
            d.calculate_stats(
                qi=['non-existent'],
                sa=['col1'],
                metrics=['k_anonymity', 'l_diversity', 't_closeness']
            )
        self.assertIn(
            "Column 'non-existent' does not exist in the given dataset",
            str(context.exception)
        )

    def test_sa_types_not_matching_sa_len(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        with self.assertRaises(AssertionError) as context:
            d.calculate_stats(
                qi=['non-existent'],
                sa=['col1', 'col2'],
                sa_types=['cat']
            )
        self.assertIn(
            "Length of sa-types doesn't equal length of sa. Either leave empty or set all data types explicitly.",
            str(context.exception)
        )

    def test_sa_types_not_valid(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        with self.assertRaises(AssertionError) as context:
            d.calculate_stats(
                qi=['non-existent'],
                sa=['col1', 'col2'],
                sa_types=['bla', 'num']
            )
        self.assertIn(
            "Wrong type given for sensitive attribute. Valid types: [cat, num]",
            str(context.exception)
        )

    def test_empty_qi(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        with self.assertRaises(AssertionError) as context:
            d.calculate_stats(
                qi=[],
                sa=['col1', 'col2'],
            )
        self.assertIn(
            "No quasi identifiers were defined. Please define at least one quasi identifier.",
            str(context.exception)
        )

    def test_unknown_metric(self):
        d = PrivacyDiagnostics("datasets/simpleset.csv")
        with self.assertRaises(AssertionError) as context:
            d.calculate_stats(
                qi=['col1'],
                sa=['col2'],
                metrics=['non-existent']
            )
        self.assertIn(
            "Unknown metrics specified. Valid types: [k_anonymity, l_diversity, t_closeness]", str(context.exception)
        )


if __name__ == '__main__':
    unittest.main()



import collections
import pandas as pd
import numpy as np
from scipy.stats import wasserstein_distance as wd
from typing import List, Dict


def k_anonymity(df: pd.DataFrame, qi: List[str]) -> pd.Series:
    return df[qi].value_counts()


def l_diversity(df: pd.DataFrame, qi: List[str], sa: List[str]) -> Dict[str, List[str]]:
    df_grouped = df.groupby(qi, as_index=False)
    return {s: sorted([len(row['unique']) for _, row in df_grouped[s].agg(['unique']).iterrows()]) for s in sa}


def t_closeness_cat(df: pd.DataFrame, qi: List[str], s: str) -> float:
    emd_max = 0
    equivalence_classes = [np.array(group['indexed_row']) for _, group in df.groupby(qi)]
    values = sorted(df[s].unique())
    m = len(values)
    p = collections.defaultdict(float, dict(df[s].value_counts() / len(df)))
    p = [p[x] for x in values]
    for ec in equivalence_classes:
        df_ec = df.iloc[ec]
        qis = collections.defaultdict(float, dict(df_ec[s].value_counts() / len(ec)))
        r = np.array([qis[x] - p[j] for j, x in enumerate(values)])
        emd_max = max(emd_max, np.sum([np.abs(r[i]) for i in range(m)]) / 2)
    return emd_max


def t_closeness_num(df: pd.DataFrame, qi: List[str], s: str) -> float:
    emd_max = 0
    equivalence_classes = [np.array(group['indexed_row']) for _, group in df.groupby(qi)]
    values = df[s]
    max_distance = max(values) - min(values)
    for ec in equivalence_classes:
        q_values = df.iloc[ec][s]
        emd_max = max(emd_max, wd(values, q_values) / max_distance)
    return emd_max


def t_closeness(df: pd.DataFrame, qi: List[str], sa: List[str], sa_types: List[str]) -> float:
    emd_max = 0
    for i, s in enumerate(sa):
        if sa_types[i] == "cat":
            emd_max = max(emd_max, t_closeness_cat(df, qi, s))
        elif sa_types[i] == "num":
            emd_max = max(emd_max, t_closeness_num(df, qi, s))
    return emd_max



from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
sns.set()


def plot(df: pd.DataFrame, k: List[str], metric: str, tmpdir: str):
    fig = plt.figure(figsize=(12, 8))
    sorted_uniques = list(sorted(np.unique(k)))
    plt.plot([min(sorted_uniques)] + sorted_uniques, [0] + list(np.cumsum([sum(x == k) * x for x in sorted_uniques]) * 100 / len(df)),
             label='unique rows loss')
    plt.plot([min(sorted_uniques)] + sorted_uniques, [0] + list(np.cumsum([sum(x == k) for x in sorted_uniques]) * 100 / len(k)),
             label='unique equivalence group loss')

    fig.suptitle(f'{metric} vs. Data Loss', fontsize=24)
    plt.xlabel(f'{metric} level', fontsize=20)
    plt.ylabel('Percentage data loss', fontsize=20)
    plt.legend(fontsize=16)
    plt.xticks(fontsize=16)
    plt.yticks(fontsize=16)
    plt.xlim(left=-10)
    plt.ylim(bottom=-10)
    plt.close(fig)
    fig.savefig(str(Path(tmpdir, f'{metric}.png')))


def plot_k_anonymity(df: pd.DataFrame, k: pd.Series, tmpdir: str):
    plot(df, list(k.values), "k-Anonymity", tmpdir)


def plot_l_diversity(df: pd.DataFrame, sa: List[str], l: dict, tmpdir: str):
    for s in sa:
        plot(df, l[s], f"l-Diversity ({s})", tmpdir)



import argparse
import logging
import tempfile

from pandas.core.dtypes.common import is_string_dtype, is_numeric_dtype

from strmprivacy.diagnostics.metrics import *
from strmprivacy.diagnostics.report.report import Report
from strmprivacy.diagnostics.plots import plot_l_diversity, plot_k_anonymity


class PrivacyDiagnostics:
    defaultMetrics = ['k_anonymity', 'l_diversity']

    def __init__(self, file: str, sample: int = 0):
        self.k_anonymity, self.l_diversity, self.t_closeness = None, None, None
        self.file = file
        self.df = pd.read_csv(file, on_bad_lines='warn')
        if 0 < sample <= len(self.df):
            self.df = self.df.sample(n=int(sample), random_state=0).reset_index(drop=True)
        self.df['indexed_row'] = self.df.index.values

    def calculate_stats(self, qi: List[str], sa: List[str] = [], sa_types=[],
                        metrics=defaultMetrics):
        self.assert_arguments(qi, sa, sa_types, metrics)
        self.assert_columns(qi + sa)

        if len(sa_types) == 0:
            for s in sa:
                if is_string_dtype(self.df[s]):
                    sa_types.append('cat')
                elif is_numeric_dtype(self.df[s]):
                    sa_types.append('num')
                else:
                    raise TypeError(f"Cannot infer type of column '{s}'")
        # k-Anonymity
        self.k_anonymity = k_anonymity(self.df, qi) if 'k_anonymity' in metrics else None
        # l-Diversity
        self.l_diversity = l_diversity(self.df, qi, sa) if 'l_diversity' in metrics and len(sa) > 0 else None
        # t-Closeness
        self.t_closeness = t_closeness(self.df, qi, sa, sa_types) if 't_closeness' in metrics and len(sa) > 0 else None

    def create_report(self, qi: List[str], sa: List[str] = [], sa_types: List[str] = [],
                      metrics=defaultMetrics, path: str = '.'):
        if ('k_anonymity' in metrics and self.k_anonymity is None) or \
                ('l_diversity' in metrics and self.l_diversity is None) or \
                ('t_closeness' in metrics and self.t_closeness is None):
            self.calculate_stats(qi, sa, sa_types, metrics)

        with tempfile.TemporaryDirectory() as tmpdir:
            # Plots
            if self.k_anonymity is not None:
                plot_k_anonymity(self.df, self.k_anonymity, tmpdir)
            if self.l_diversity is not None:
                plot_l_diversity(self.df, sa, self.l_diversity, tmpdir)

            # Report
            Report(tmpdir=tmpdir).create(self.k_anonymity, self.l_diversity, self.t_closeness, path)

    def assert_columns(self, columns: List[str]):
        for column in columns:
            assert column in self.df.columns, f"Column '{column}' does not exist in the given dataset"

    @staticmethod
    def assert_arguments(qi: List[str], sa: List[str], sa_types: List[str] = [],
                         metrics: List[str] = defaultMetrics):
        if len(sa_types) > 0:
            assert len(sa_types) == len(sa), "Length of sa-types doesn't equal length of sa. " \
                                             "Either leave empty or set all data types explicitly."
            assert all(x in ["cat", "num"] for x in sa_types), "Wrong type given for sensitive attribute. " \
                                                               "Valid types: [cat, num]"
        else:
            logging.warning("No sensitive attribute type was defined. Types will be inferred.")
        assert len(qi), "No quasi identifiers were defined. Please define at least one quasi identifier."
        if len(sa) == 0:
            logging.warning("No sensitive attributes were defined. Only k-anonymity can be calculated.")
        assert not any(x in qi for x in sa), "A quasi identifier cannot also be a sensitive attribute."
        assert all(m in ['k_anonymity', 'l_diversity', 't_closeness'] for m in metrics), \
            "Unknown metrics specified. Valid types: [k_anonymity, l_diversity, t_closeness]"
        if 't_closeness' in metrics:
            logging.warning("t-Closeness can drastically increase calculation times with larger datasets and number "
                            "of sensitive attributes.")

    @property
    def stats(self):
        return {
            'k': self.k if self.k_anonymity is not None else 0,
            'l': self.l if self.l_diversity is not None else 0,
            't': self.t if self.t_closeness is not None else 0
        }

    @property
    def k(self):
        return min(self.k_anonymity)

    @property
    def l(self):
        return {k: min(v) for k, v in self.l_diversity.items()}

    @property
    def t(self):
        return self.t_closeness


def main(args: dict):
    file, qi, sa, sa_types, sample, report_path, metrics = args['file'], args['qi'], args['sa'], args['sa_types'], args[
        'sample'], args['report_path'], args['metrics']
    diagnostics = PrivacyDiagnostics(file, sample)
    diagnostics.assert_arguments(qi, sa, sa_types, metrics)
    diagnostics.create_report(qi, sa, sa_types, metrics, report_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', type=str, help='path to data file')
    parser.add_argument('--qi', nargs='+', help='names of the quasi identifier columns', default=[])
    parser.add_argument('--sa', nargs='+', help='names of the sensitive attribute columns', default=[])
    parser.add_argument('--sa-types', nargs='+', help='data types of the sensitive attribute columns. '
                                                      'valid types: [cat, num]', default=[])
    parser.add_argument('--metrics', nargs='+', help='pick the metric you want in your report',
                        default=['k_anonymity', 'l_diversity'])
    parser.add_argument('--sample', type=int, help='random sample size', default=0)
    parser.add_argument('--report-path', type=str, help='path to save report to', default='.')
    arguments = vars(parser.parse_args())
    main(arguments)



import logging
from pathlib import Path

from fpdf import FPDF


class Report(FPDF):
    def __init__(self, orientation='P', unit='mm', format='A4', tmpdir=""):
        super().__init__(orientation, unit, format)
        self.tmpdir = tmpdir

    def header(self):
        # Arial bold 15
        try:
            self.set_font('Inter', size=15)
        except:
            self.set_font('arial', size=15)
        # Move to the right
        self.cell(80)
        # Title
        self.cell(30, 10, 'STRM Privacy Report', 0, 0, 'C')
        # Line break
        self.ln(20)

    def intro(self):
        try:
            self.set_font('Inter', style="i", size=10)
        except:
            self.set_font('arial', style="i", size=10)
        self.write(h=5, txt=f"Using this report, you can assess if your privacy transformations are fit-for-purpose."
                            " K-anonymity, and l-diversity are methods to assess the degree of anonymity in a given dataset. T-closeness indicates the information loss in your data, and so its utility."
                            " The methods look at how unique a combination of datapoints is, giving you insight into the risk of re-identification: the less unique, the lower the probability you can find a single individual in the data."
                            " Therefore, generally speaking, higher scores are better: they indicate it's harder to find a single individual in your data."
                   )
        self.ln(5)

    def summation(self, k, l, t):
        try:
            self.set_font('Inter', size=8)
        except:
            self.set_font('arial', size=8)

        i = 1
        if k is not None:
            self.write(h=5, txt=f"{i}. Your data has a k-Anonymity of {min(k)}")
            self.ln(5)
            i += 1
        if l is not None:
            self.write(h=5, txt=f"{i}.  l-Diversity minimal distinct values per equivalence group: " +
                                ', '.join([f"{key}: {min(value)}" for key, value in l.items()]))
            self.ln(5)
            i += 1
        if t is not None:
            self.write(h=5, txt=f"{i}. Your data has a t-Closeness of {t:.4f}")
            self.ln(5)

    def plots(self, metric):
        width = self.w * 0.7
        height = width * 2 / 3
        x = (self.w - width) // 2
        y = self.y
        if y + height > self.h:
            self.add_page()
        self.image(str(Path(self.tmpdir, f'{metric}.png')), x=x, w=width, h=height)

    def k_metrics(self, k):
        try:
            self.set_font('Inter', size=12)
        except:
            self.set_font('arial', size=12)
        self.write(h=5, txt="k-Anonymity")
        self.ln(5)
        try:
            self.set_font('Inter', size=8)
        except:
            self.set_font('arial', size=8)
        self.write(h=5, txt=f'Your data has a k-Anonymity of  {min(k)}')
        self.ln(10)
        self.write(
            h=5,
            txt="K-anonymity indicates if you can isolate a given data subject, based on how many equal groups can be found in the data. "
                "Any K of or above 2 indicates at least two indivduals belong to any group inside the data."
                "The graph below should be read as follows. The unique rows line (blue) indicates what percentage of "
                "the total rows (y-axis) you will lose to achieve every k-Anonymity level (x-axis), "
                "without transforming data. The loss in unique equivalence groups line (orange) indicates the "
                "percentage of unique combinations of the values in the quasi identifier columns (y-axis) you will "
                "lose to achieve every k-Anonymity level (x-axis)."
        )
        self.ln(10)
        self.plots('k-Anonymity')
        self.ln(15)

    def l_metrics(self, l):
        try:
            self.set_font('Inter', size=12)
        except:
            self.set_font('arial', size=12)
        self.write(h=5, txt="l-Diversity")
        self.ln(5)
        try:
            self.set_font('Inter', size=8)
        except:
            self.set_font('arial', size=8)
        self.write(h=5, txt="Minimal distinct values per equivalence group:")
        self.ln(5)
        for k, v in l.items():
            self.write(h=5, txt=f'{k}: {min(v)}')
            self.ln(5)
        self.ln(5)
        self.write(
            h=5,
            txt="L-diversity extends k-anonymity by also looking at every attribute in each equivalence class (the group of equal attributes in your data). "
                "A higher L indicates higher variability in each class, so that attempts for re-identification always yield sufficient uncertainty on the exact individual within that class. "
                "The graphs below should be read as follows. The unique rows line (blue) indicates what percentage of "
                "the total rows (y-axis) you will lose to achieve every l-Diversity level (x-axis) for each sensitive "
                "attribute individually. The loss in unique equivalence groups line (orange) indicates the percentage "
                "of unique combinations of the values in the quasi identifier columns (y-axis) you will lose to "
                "achieve every l-Diversity level (x-axis) for each sensitive attribute. "
        )
        self.ln(10)
        for k, v in l.items():
            self.plots(f'l-Diversity ({k})')
            self.ln(10)

    def t_metrics(self, t):
        try:
            self.set_font('Inter', size=12)
        except:
            self.set_font('arial', size=12)
        self.write(h=5, txt="t-Closeness")
        self.ln(5)
        try:
            self.set_font('Inter', size=8)
        except:
            self.set_font('arial', size=8)
        self.write(h=5,
                   txt="T-closeness indicates how close the data in each L-diverse group is to the original data, and so how much utility is retained inside the data. "
                       " A T closer to 1 indicates the data is (statistically) closer to the input data.")
        self.ln(5)
        self.write(h=5, txt=f"Your data has a t-closeness of  {t:.4f}")
        self.ln(15)

    def create(self, k, l, t, path='.'):
        try:
            self.add_font('Inter', fname='strmprivacy/diagnostics/report/Inter-V.ttf', uni=True)
        except RuntimeError:
            logging.info('Inter not found, using arial')
        finally:
            self.accept_page_break()
            self.add_page()
            self.summation(k, l, t)
            if k is not None:
                self.k_metrics(k)
            if l is not None:
                self.l_metrics(l)
            if t is not None:
                self.t_metrics(t)
            Path(path, 'report.py')
            file_path = Path(path, 'STRM-Privacy-report.pdf')
            self.output(str(file_path), 'F')
            print(f"Report saved to {file_path}")



#!/usr/bin/env python
from setuptools import setup, find_namespace_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    author="Stream Machine B.V.",
    author_email='apis@strmprivacy.io',
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    description="STRM Privacy Diagnostics for reporting privacy metrics on a dataset.",
    install_requires=[
        "pandas>=1.4.3",
        "numpy>=1.23.2",
        "matplotlib>=3.5.3",
        "fpdf>=1.7.2",
        "temp>=2020.7.2",
        "argparse>=1.4.0",
        "pathlib>=1.0.1",
        "seaborn>=0.11.2",
        "setuptools~=58.0.4",
        "scipy~=1.9.1"
    ],
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='strmprivacy privacy diagnostics',
    name='strmprivacy-diagnostics',
    packages=find_namespace_packages(include=['strmprivacy.*']),
    namespace_packages=["strmprivacy"],
    version='1.2.2',
    zip_safe=False,
)
